# flake8: noqa
from dataclasses import fields
from .containers import (
    TYPES,
    Locator,
    PointLocator,
    OffsetLocator,
    RegionLocator,
    SizeLocator,
    ImageLocator,
    OcrLocator,
    BrowserLocator,
    WindowsLocator,
    # For backwards compatibility:
    Coordinates,
    Offset,
    BrowserDOM,
    ImageTemplate,
)
from .database import LocatorsDatabase, sanitize_name
from .literal import LocatorType



import re
from enum import Enum
from typing import Callable, List, NamedTuple, Optional, Union

from RPA.core.geometry import Point, Region, Undefined
from RPA.core.locators import literal, Locator


Geometry = Union[Point, Region, Undefined]


class InvalidSyntax(ValueError):
    """Raised when given syntax is malformed."""


class Token(Enum):
    """Possible tokens in a locator."""

    THEN = "then"
    AND = "and"
    OR = "or"
    NOT = "not"
    LPAREN = "left parenthesis"
    RPAREN = "right parenthesis"
    LOCATOR = "locator"


ValueType = Union[Locator, "Not", "Expression", "Chain"]


class TokenPair(NamedTuple):
    """Pair of resolved token and original value."""

    token: Token
    value: Union[str, Locator]


class Not(NamedTuple):
    """Negated value."""

    value: ValueType

    def __repr__(self):
        return f"Not({self.value})"


class Expression(NamedTuple):
    """A conditional expression between two values."""

    lhs: ValueType
    op: Token
    rhs: ValueType

    def __repr__(self):
        return f"Expression({self.lhs} {self.op.value} {self.rhs})"


class Chain(tuple):
    """A set of consecutive values."""

    def __new__(cls, *args):
        return super().__new__(cls, args)

    def __repr__(self):
        return f"Chain({', '.join(str(val) for val in self)})"


class Peekable:
    """Generator-like wrapper with the ability
    to inspect the next value without consuming it.
    """

    EOF = object()

    def __init__(self, iterable):
        self._iterable = iter(iterable)
        self._current = self.EOF
        self._peek = self.EOF
        self._next()

    def __iter__(self):
        return self

    def __next__(self):
        if self.is_empty:
            raise StopIteration

        self._next()
        return self._current

    @property
    def current(self):
        return self._current

    @property
    def peek(self):
        return self._peek

    @property
    def is_empty(self):
        return self._peek is self.EOF

    def _next(self):
        self._current = self._peek
        try:
            self._peek = next(self._iterable)
        except StopIteration:
            self._peek = self.EOF


class Tokenizer:
    """Methods for tokenizing a locator string."""

    # fmt: off
    TOKEN_PATTERNS = [
        r"(?P<THEN>then|\+)",
        r"(?P<AND>and|\&\&|\&)",
        r"(?P<OR>or|\|\||\|)",
        r"(?P<NOT>not|\!)",
        r"(?P<LPAREN>\()",
        r"(?P<RPAREN>\))",
        (
            r"(?P<LOCATOR>"
            r"(\w+\:)?"      # Optional locator typename, e.g. ocr:
            r"(("            # Argument definition:
            r"\"[^\"]*\""    # 1. If quoted, continue until end quote
            r"|"             # or
            r"[^\s\)]+"      # 2. If unquoted, continue until whitespace or paren
            r")[\,]?)+"      # One or more arguments, comma delimited
            r")"
        ),
        r"(?P<SKIP>\s+)",
        r"(?P<MISMATCH>.+)",
    ]
    # fmt: on

    PATTERN = re.compile("|".join(TOKEN_PATTERNS))

    @classmethod
    def tokenize(
        cls, locator: str, parser: Callable[[str], Locator]
    ) -> List[TokenPair]:
        """Convert locator string to list of token pairs."""
        scanner = cls.PATTERN.scanner(locator)

        tokens = []
        for match in iter(scanner.match, None):
            name = match.lastgroup
            value = match.group()

            if name == "MISMATCH":
                raise InvalidSyntax(f"Unknown token: '{value}'")

            if name == "SKIP":
                continue

            if name == "LOCATOR":
                value = parser(value)

            pair = TokenPair(Token[name], value)
            tokens.append(pair)

        if not tokens:
            raise InvalidSyntax("Empty expression")

        return tokens


class SyntaxParser:
    """Methods for parsing a locator string into a tree."""

    def __init__(self):
        self.tokens = Peekable([])

    def parse(self, locator: str, parser: Callable[[str], Locator]) -> ValueType:
        """Parse locator string to tokenized expression."""
        pairs = Tokenizer.tokenize(locator, parser)
        self.tokens = Peekable(pairs)
        return self._chain()

    def _chain(self) -> ValueType:
        """Chain of expressions."""
        links = [self._expression()]
        while self._accept(Token.THEN):
            links.append(self._expression())

        if len(links) == 1:
            return links[0]

        return Chain(*links)

    def _expression(self) -> ValueType:
        """Either a single value or expression between two values."""
        expr = self._value()
        while self._accept(Token.AND) or self._accept(Token.OR):
            op = self.tokens.current
            rhs = self._value()
            expr = Expression(expr, op.token, rhs)

        return expr

    def _value(self) -> ValueType:
        """Either a locator literal or nested expression inside parentheses."""
        if self._accept(Token.NOT):
            return Not(self._value())
        elif self._accept(Token.LOCATOR):
            return self.tokens.current.value
        elif self._accept(Token.LPAREN):
            chain = self._chain()
            self._expect(Token.RPAREN)
            return chain
        else:
            raise InvalidSyntax("Expected locator or parentheses")

    def _consume(self) -> TokenPair:
        """Step forward to the next token and return the current one."""
        try:
            return next(self.tokens)
        except StopIteration:
            # pylint: disable=raise-missing-from
            raise InvalidSyntax("Unexpected end of expression")

    def _accept(self, token: Token) -> bool:
        """Consume tokens if the next one matches."""
        if not self.tokens.is_empty and self.tokens.peek.token == token:
            self._consume()
            return True
        else:
            return False

    def _expect(self, token: Token):
        """Assert that the next token is as expected."""
        if self.tokens.is_empty:
            raise InvalidSyntax("Unexpected end of expression")

        if not self._accept(token):
            actual = self.tokens.peek.token
            raise InvalidSyntax(f"Expected '{token.value}', was '{actual.value}'")


class Resolver:
    """Parser for locator expressions and resolving them into a set of final values."""

    def __init__(self, finder, database: Optional[str] = None):
        self.finder = finder
        self.database = database
        self._stack = [Undefined()]

    @property
    def current(self):
        return self._stack[-1]

    def dispatch(self, locator: str):
        """Visit locator tree."""

        def from_literal(value: str) -> Locator:
            return literal.parse(value, self.database)

        root = SyntaxParser().parse(locator=locator, parser=from_literal)
        return sorted(self._resolve(root))

    def _resolve(self, value: Union[ValueType, bool]):
        """Resolve value of any type."""
        resolvers = {
            Locator: self._locator,
            Not: self._not,
            Expression: self._expression,
            Chain: self._chain,
        }

        for type_, func in resolvers.items():
            if isinstance(value, type_):
                return func(value)

        raise RuntimeError(f"Unexpected value: {value}")

    def _chain(self, chain: Chain):
        """Resolve chain of expressions."""
        current = self._resolve(chain[0])

        for link in chain[1:]:
            temp = []
            for base in current:
                self._stack.append(base)
                try:
                    value = self._resolve(link)
                    if not value:
                        continue
                    temp.extend(value)
                finally:
                    self._stack.pop()

            current = []
            for value in temp:
                if value not in current:
                    current.append(value)

        return current

    def _expression(self, expr: Expression):
        """Resolve conditional expression."""
        if expr.op == Token.OR:
            lhs = self._resolve(expr.lhs)
            if lhs:
                return lhs
            rhs = self._resolve(expr.rhs)
            if rhs:
                return rhs
            return []

        elif expr.op == Token.AND:
            lhs = self._resolve(expr.lhs)
            if not lhs:
                return []
            rhs = self._resolve(expr.rhs)
            if not rhs:
                return []
            return lhs + rhs

        else:
            raise RuntimeError(f"Unexpected expression: {expr}")

    def _locator(self, locator: Locator):
        """Resolve final values for locator."""
        return self.finder(self.current, locator)

    def _not(self, negated: Not):
        """Negate resolved values."""
        value = self._resolve(negated.value)
        if not value:
            return [Undefined()]
        else:
            return []



# Copyright 2017- Robot Framework Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generic test library core for Robot Framework.

Main usage is easing creating larger test libraries. For more information and
examples see the project pages at
https://github.com/robotframework/PythonLibCore
"""

import inspect
import os
import sys
import typing


# Explicitly set here instead of importing stuff from `robot`.
PY_VERSION = sys.version_info[:3]
robot_version = "4."
RF32 = robot_version < "4."  # `False` all the time

__version__ = "3.0.0"


# pylint: disable=missing-class-docstring
class HybridCore:
    def __init__(self, library_components):
        self.keywords = {}
        self.keywords_spec = {}
        self.attributes = {}
        self.add_library_components(library_components)
        self.add_library_components([self])

    def add_library_components(self, library_components):
        self.keywords_spec["__init__"] = KeywordBuilder.build(self.__init__)
        for component in library_components:
            for name, func in self.__get_members(component):
                if callable(func) and hasattr(func, "robot_name"):
                    kw = getattr(component, name)
                    kw_name = func.robot_name or name
                    self.keywords[kw_name] = kw
                    self.keywords_spec[kw_name] = KeywordBuilder.build(kw)
                    # Expose keywords as attributes both using original
                    # method names as well as possible custom names.
                    self.attributes[name] = self.attributes[kw_name] = kw

    def __get_members(self, component):
        if inspect.ismodule(component):
            return inspect.getmembers(component)
        if inspect.isclass(component):
            raise TypeError(
                "Libraries must be modules or instances, got "
                "class {!r} instead.".format(component.__name__)
            )
        # pylint: disable=unidiomatic-typecheck
        if type(component) != component.__class__:
            raise TypeError(
                "Libraries must be modules or new-style class "
                "instances, got old-style class {!r} instead.".format(
                    component.__class__.__name__
                )
            )
        return self.__get_members_from_instance(component)

    def __get_members_from_instance(self, instance):
        # Avoid calling properties by getting members from class, not instance.
        cls = type(instance)
        for name in dir(instance):
            owner = cls if hasattr(cls, name) else instance
            yield name, getattr(owner, name)

    def __getattr__(self, name):
        if name in self.attributes:
            return self.attributes[name]
        raise AttributeError(
            "{!r} object has no attribute {!r}".format(type(self).__name__, name)
        )

    def __dir__(self):
        my_attrs = super().__dir__()
        return sorted(set(my_attrs) | set(self.attributes))

    def get_keyword_names(self):
        return sorted(self.keywords)


class DynamicCore(HybridCore):
    def run_keyword(self, name, args, kwargs=None):
        return self.keywords[name](*args, **(kwargs or {}))

    def get_keyword_arguments(self, name):
        spec = self.keywords_spec.get(name)
        return spec.argument_specification

    def get_keyword_tags(self, name):
        return self.keywords[name].robot_tags

    def get_keyword_documentation(self, name):
        if name == "__intro__":
            return inspect.getdoc(self) or ""
        spec = self.keywords_spec.get(name)
        return spec.documentation

    def get_keyword_types(self, name):
        spec = self.keywords_spec.get(name)
        if spec is None:
            raise ValueError('Keyword "%s" not found.' % name)
        return spec.argument_types

    def __get_keyword(self, keyword_name):
        if keyword_name == "__init__":
            return self.__init__
        if keyword_name.startswith("__") and keyword_name.endswith("__"):
            return None
        method = self.keywords.get(keyword_name)
        if not method:
            raise ValueError('Keyword "%s" not found.' % keyword_name)
        return method

    def get_keyword_source(self, keyword_name):
        method = self.__get_keyword(keyword_name)
        path = self.__get_keyword_path(method)
        line_number = self.__get_keyword_line(method)
        if path and line_number:
            return "{}:{}".format(path, line_number)
        if path:
            return path
        if line_number:
            return ":%s" % line_number
        return None

    def __get_keyword_line(self, method):
        try:
            lines, line_number = inspect.getsourcelines(method)
        except (OSError, TypeError):
            return None
        for increment, line in enumerate(lines):
            if line.strip().startswith("def "):
                return line_number + increment
        return line_number

    def __get_keyword_path(self, method):
        try:
            return os.path.normpath(inspect.getfile(method))
        except TypeError:
            return None


class KeywordBuilder:
    @classmethod
    def build(cls, function):
        return KeywordSpecification(
            argument_specification=cls._get_arguments(function),
            documentation=inspect.getdoc(function) or "",
            argument_types=cls._get_types(function),
        )

    @classmethod
    def unwrap(cls, function):
        return inspect.unwrap(function)

    @classmethod
    def _get_arguments(cls, function):
        unwrap_function = cls.unwrap(function)
        arg_spec = cls._get_arg_spec(unwrap_function)
        argument_specification = cls._get_default_and_named_args(arg_spec, function)
        argument_specification.extend(cls._get_var_args(arg_spec))
        kw_only_args = cls._get_kw_only(arg_spec)
        if kw_only_args:
            argument_specification.extend(kw_only_args)
        argument_specification.extend(cls._get_kwargs(arg_spec))
        return argument_specification

    @classmethod
    def _get_arg_spec(cls, function):
        return inspect.getfullargspec(function)

    @classmethod
    def _get_default_and_named_args(cls, arg_spec, function):
        args = cls._drop_self_from_args(function, arg_spec)
        args.reverse()
        defaults = list(arg_spec.defaults) if arg_spec.defaults else []
        formated_args = []
        for arg in args:
            if defaults:
                formated_args.append((arg, defaults.pop()))
            else:
                formated_args.append(arg)
        formated_args.reverse()
        return formated_args

    @classmethod
    def _drop_self_from_args(cls, function, arg_spec):
        return arg_spec.args[1:] if inspect.ismethod(function) else arg_spec.args

    @classmethod
    def _get_var_args(cls, arg_spec):
        if arg_spec.varargs:
            return ["*%s" % arg_spec.varargs]
        return []

    @classmethod
    def _get_kwargs(cls, arg_spec):
        return ["**%s" % arg_spec.varkw] if arg_spec.varkw else []

    @classmethod
    def _get_kw_only(cls, arg_spec):
        kw_only_args = []
        for arg in arg_spec.kwonlyargs:
            if not arg_spec.kwonlydefaults or arg not in arg_spec.kwonlydefaults:
                kw_only_args.append(arg)
            else:
                value = arg_spec.kwonlydefaults.get(arg, "")
                kw_only_args.append((arg, value))
        return kw_only_args

    @classmethod
    def _get_types(cls, function):
        if function is None:
            return function
        types = getattr(function, "robot_types", ())
        if types is None or types:
            return types
        return cls._get_typing_hints(function)

    @classmethod
    def _get_typing_hints(cls, function):
        function = cls.unwrap(function)
        try:
            hints = typing.get_type_hints(function)
        except Exception:  # pylint: disable=broad-except
            hints = function.__annotations__
        arg_spec = cls._get_arg_spec(function)
        all_args = cls._args_as_list(function, arg_spec)
        for arg_with_hint in list(hints):
            # remove return and self statements
            if arg_with_hint not in all_args:
                hints.pop(arg_with_hint)
        if RF32:
            default = cls._get_defaults(arg_spec)
            return cls._remove_optional_none_type_hints(hints, default)
        return hints

    @classmethod
    def _args_as_list(cls, function, arg_spec):
        function_args = []
        function_args.extend(cls._drop_self_from_args(function, arg_spec))
        if arg_spec.varargs:
            function_args.append(arg_spec.varargs)
        function_args.extend(arg_spec.kwonlyargs or [])
        if arg_spec.varkw:
            function_args.append(arg_spec.varkw)
        return function_args

    # TODO: Remove when support RF 3.2 is dropped
    # Copied from: robot.running.arguments.argumentparser
    @classmethod
    def _remove_optional_none_type_hints(cls, type_hints, defaults):
        # If argument has None as a default, typing.get_type_hints adds
        # optional None to the information it returns. We don't want that.
        for arg, default in defaults:
            if default is None and arg in type_hints:
                type_ = type_hints[arg]
                if cls._is_union(type_):
                    types = type_.__args__
                    if len(types) == 2 and types[1] is type(None):  # noqa
                        type_hints[arg] = types[0]
        return type_hints

    # TODO: Remove when support RF 3.2 is dropped
    # Copied from: robot.running.arguments.argumentparser
    @classmethod
    def _is_union(cls, typing_type):
        if PY_VERSION >= (3, 7) and hasattr(typing_type, "__origin__"):
            typing_type = typing_type.__origin__
        return isinstance(typing_type, type(typing.Union))

    @classmethod
    def _get_defaults(cls, arg_spec):
        if not arg_spec.defaults:
            return {}
        names = arg_spec.args[-len(arg_spec.defaults) :]
        return zip(names, arg_spec.defaults)


class KeywordSpecification:
    def __init__(
        self, argument_specification=None, documentation=None, argument_types=None
    ):
        self.argument_specification = argument_specification
        self.documentation = documentation
        self.argument_types = argument_types



import os
import platform
from pathlib import Path


def robocorp_home():
    """Get the absolute path to the user's Robocorp home folder.
    Prefers environment variable ROBOCORP_HOME, if defined.
    """
    env = os.getenv("ROBOCORP_HOME", "")

    if env.strip():
        path = Path(env)
    elif platform.system() == "Windows":
        path = Path.home() / "AppData" / "Local" / "robocorp"
    else:
        path = Path.home() / ".robocorp"

    return path.resolve()



from collections.abc import Iterable


def is_dict_like(obj):
    """Check if `obj` behaves like a dictionary."""
    return all(
        hasattr(obj, attr) for attr in ("__getitem__", "keys", "__contains__")
    ) and not isinstance(obj, type)


def is_list_like(obj):
    """Check if `obj` behaves like a list."""
    return isinstance(obj, Iterable) and not isinstance(obj, (str, bytes))


def is_namedtuple(obj):
    """Check if `obj` is a namedtuple."""
    return isinstance(obj, tuple) and hasattr(obj, "_fields")



import platform
from typing import Dict

IS_WINDOWS = platform.system() == "Windows"

if IS_WINDOWS:
    import psutil


def get_process_list() -> Dict:
    """Get process list.

    Returns dictionary mapping process id to process name
    """
    return {proc.pid: proc.name() for proc in psutil.process_iter()}



from pathlib import Path
from typing import Union

from RPA.core.vendor.deco import keyword as method
from RPA.core.windows.context import ActionNotPossible, WindowsContext
from RPA.core.windows.locators import Locator


class ActionMethods(WindowsContext):
    """Keywords for performing desktop actions."""

    @method
    def screenshot(self, locator: Locator, filename: Union[str, Path]) -> str:
        # Saves the control image from the provided element.
        item = self.ctx.get_element(locator).item
        meth = "CaptureToImage"
        capture_to_image = getattr(item, meth, None)
        if not capture_to_image:
            raise ActionNotPossible(
                f"Element found with {locator!r} does not have the {meth!r}"
                " attribute"
            )

        item.SetFocus()
        filepath = str(Path(filename).expanduser().resolve())
        capture_to_image(filepath)
        return filepath



import contextlib
import functools
import logging
import os
import platform
import stat
from pathlib import Path
from typing import Dict, List, Optional

from packaging import version as version_parser
from selenium import webdriver
from selenium.webdriver.common.service import Service
from selenium.webdriver.remote.webdriver import WebDriver
from webdriver_manager.chrome import ChromeDriverManager as _ChromeDriverManager
from webdriver_manager.core.download_manager import DownloadManager
from webdriver_manager.core.driver_cache import DriverCacheManager
from webdriver_manager.core.logger import log
from webdriver_manager.core.manager import DriverManager
from webdriver_manager.core.os_manager import ChromeType, OperationSystemManager
from webdriver_manager.drivers.chrome import ChromeDriver as _ChromeDriver
from webdriver_manager.firefox import GeckoDriverManager
from webdriver_manager.microsoft import (
    EdgeChromiumDriverManager,
    IEDriverManager as _IEDriverManager,
)
from webdriver_manager.opera import OperaDriverManager

from RPA.core.robocorp import robocorp_home


class ChromeDriver(_ChromeDriver):
    """Custom class which correctly obtains the chromedriver download URL."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._resolve_version = False

    def _get_resolved_version(self, determined_browser_version: Optional[str]) -> str:
        if determined_browser_version:
            # Resolution works with the first 3 atoms of the version, so exclude the
            #  4th one in case it exists.
            determined_browser_version = ".".join(
                determined_browser_version.split(".")[:3]
            )

        latest_release_url = (
            self._latest_release_url
            if (
                self._driver_version_to_download == "latest"
                or determined_browser_version is None
            )
            else f"{self._latest_release_url}_{determined_browser_version}"
        )
        resp = self._http_client.get(url=latest_release_url)
        return resp.text.rstrip()

    def get_latest_release_version(self) -> str:
        # This is activated for any chromedriver version.
        determined_browser_version = self.get_browser_version_from_os()
        if determined_browser_version and not self._resolve_version:
            parts = version_parser.parse(determined_browser_version).release
            if len(parts) == 4:
                # Got a fully downloadable version that MAY be available, but we are
                #  not sure until we don't try it.
                log(
                    f"Get {determined_browser_version} {self._name} version for"
                    f" {self._browser_type}"
                )
                return determined_browser_version

        log(
            f"Get LATEST {self._name} version for {self._browser_type} based on"
            f" {determined_browser_version}"
        )
        return self._get_resolved_version(determined_browser_version)

    def _resolve_modern_url(
        self, version: Dict, *, driver_platform: str
    ) -> Optional[str]:
        downloads = version["downloads"]["chromedriver"]
        for dld in downloads:
            # There's also a chance for platform mismatch.
            if dld["platform"] == driver_platform:
                # Ensure the caching key and driver path matches the exact resolved
                #  version.
                self._driver_version_to_download = version["version"]
                return dld["url"]

        return None

    # pylint: disable=arguments-renamed
    def get_url_for_version_and_platform(
        self, browser_version: str, driver_platform: str
    ) -> str:
        # This is activated for chromedriver 115 or higher only.
        parse_version = lambda ver: version_parser.parse(ver).release  # noqa: E731
        parse_floating_version = lambda ver: parse_version(ver)[:3]  # noqa: E731
        parsed_floating_version = parse_floating_version(browser_version)
        resolve_modern_url = functools.partial(
            self._resolve_modern_url, driver_platform=driver_platform
        )

        versions_url = (
            "https://googlechromelabs.github.io/chrome-for-testing/"
            "known-good-versions-with-downloads.json"
        )
        response = self._http_client.get(versions_url)
        data = response.json()
        versions = data["versions"]

        candidates = []
        for version in versions:
            if version["version"] == browser_version:  # exact match
                url = resolve_modern_url(version)
                if url:
                    return url
            elif parsed_floating_version == parse_floating_version(version["version"]):
                candidates.append(version)

        # No exact version found, let's return the latest from the matching candidates.
        if candidates:
            latest = max(
                candidates, key=lambda candidate: parse_version(candidate["version"])
            )
            url = resolve_modern_url(latest)
            if url:
                return url

        raise Exception(f"No such driver version {browser_version} for {platform}")

    # pylint: disable=arguments-differ
    def get_driver_download_url(self, os_type: str, resolve: bool = False) -> str:
        if resolve:
            # This time we want a resolved version based on the previously parsed
            #  non-existing one on the server.
            self._resolve_version = True  # force resolve
            self._driver_version_to_download = None  # resets cache

        return super().get_driver_download_url(os_type)


class ChromeDriverManager(_ChromeDriverManager):
    """Custom Chrome webdriver manager which correctly downloads the chromedriver."""

    def __init__(
        self,
        driver_version: Optional[str] = None,
        name: str = "chromedriver",
        url: str = "https://chromedriver.storage.googleapis.com",
        latest_release_url: str = (
            "https://chromedriver.storage.googleapis.com/LATEST_RELEASE"
        ),
        chrome_type: str = ChromeType.GOOGLE,
        download_manager: Optional[DownloadManager] = None,
        cache_manager: Optional[DriverCacheManager] = None,
        os_system_manager: Optional[OperationSystemManager] = None,
    ):
        super().__init__(
            driver_version=driver_version,
            name=name,
            url=url,
            latest_release_url=latest_release_url,
            chrome_type=chrome_type,
            download_manager=download_manager,
            cache_manager=cache_manager,
            os_system_manager=os_system_manager,
        )

        # Replace the upstream webdriver helper class with our custom auto-resolving
        #  behavior.
        self.driver = ChromeDriver(
            name=name,
            driver_version=driver_version,
            url=url,
            latest_release_url=latest_release_url,
            chrome_type=chrome_type,
            http_client=self.http_client,
            os_system_manager=os_system_manager,
        )

    def _get_driver_binary_path(self, driver: ChromeDriver) -> str:
        """Resolve the webdriver version if is not available for download."""
        try:
            return super()._get_driver_binary_path(driver)
        except ValueError:  # parsed version isn't found, let's try to resolve it
            os_type = self.get_os_type()
            resolved_url = driver.get_driver_download_url(os_type, resolve=True)
            file = self._download_manager.download_file(resolved_url)
            binary_path = self._cache_manager.save_file_to_cache(driver, file)
            return binary_path


class IEDriverManager(_IEDriverManager):
    """Custom driver manager class for IE webdriver. (running Edge in IE mode)"""

    # Forcefully download the 32bit version of the webdriver no matter the architecture
    #  of the Windows system since it is known that the 64bit version is limited and
    #  also creates issues, like freezing browser automation with Selenium.
    # https://www.selenium.dev/documentation/webdriver/browsers/internet_explorer/
    FORCE_32BIT = True

    def get_os_type(self) -> str:
        return "Win32" if self.FORCE_32BIT else super().get_os_type()


LOGGER = logging.getLogger(__name__)

DRIVER_ROOT = robocorp_home() / "webdrivers"
DRIVER_ROOT.mkdir(parents=True, exist_ok=True)
AVAILABLE_DRIVERS = {
    # Driver names taken from `webdrivermanager` and adapted to `webdriver_manager`.
    "chrome": ChromeDriverManager,
    "firefox": GeckoDriverManager,
    "gecko": GeckoDriverManager,
    "mozilla": GeckoDriverManager,
    # NOTE: Selenium 4 dropped support for Opera.
    #  (https://github.com/SeleniumHQ/selenium/issues/10835)
    "opera": OperaDriverManager,
    # NOTE: In Selenium 4 `Edge` is the same with `ChromiumEdge`.
    "edge": EdgeChromiumDriverManager,
    "chromiumedge": EdgeChromiumDriverManager,
    # NOTE: IE is discontinued and not supported/encouraged anymore.
    "ie": IEDriverManager,
}
# Available `WebDriver` classes in Selenium.
SUPPORTED_BROWSERS = dict(
    {name: name.capitalize() for name in AVAILABLE_DRIVERS},
    **{"chromiumedge": "ChromiumEdge"},
)
_DRIVER_PREFERENCE = {
    "Windows": ["Chrome", "Firefox", "Edge"],
    "Linux": ["Chrome", "Firefox", "Edge"],
    "Darwin": ["Chrome", "Firefox", "Edge", "Safari"],
    "default": ["Chrome", "Firefox"],
}

_OPS_MANAGER = OperationSystemManager()


def _get_browser_order_from_env() -> Optional[List[str]]:
    browsers: str = os.getenv("RPA_SELENIUM_BROWSER_ORDER", "")
    if browsers:
        return [browser.strip() for browser in browsers.split(sep=",")]

    return None  # meaning there's no env var to control the order


def get_browser_order() -> List[str]:
    """Get a list of preferred browsers based on the environment variable
    `RPA_SELENIUM_BROWSER_ORDER` if set.

    The OS dictates the order if no such env var is set.
    """
    browsers: Optional[List[str]] = _get_browser_order_from_env()
    if browsers:
        return browsers

    return _DRIVER_PREFERENCE.get(platform.system(), _DRIVER_PREFERENCE["default"])


@contextlib.contextmanager
def suppress_logging():
    """Suppress webdriver-manager logging."""
    wdm_log = "WDM_LOG"
    original_value = os.getenv(wdm_log, "")
    try:
        os.environ[wdm_log] = str(logging.NOTSET)
        yield
    finally:
        os.environ[wdm_log] = original_value


def start(browser: str, service: Optional[Service] = None, **options) -> WebDriver:
    """Start a webdriver with the given options."""
    browser = browser.strip()
    webdriver_factory = getattr(webdriver, browser, None)
    if not webdriver_factory:
        raise ValueError(f"Unsupported Selenium browser: {browser}")

    # NOTE: It is recommended to pass a `service` rather than deprecated `options`.
    driver = webdriver_factory(service=service, **options)
    return driver


def _is_chromium() -> bool:
    """Detects if Chromium is used instead of Chrome no matter the platform."""
    is_browser = lambda browser_type: bool(  # noqa: E731
        _OPS_MANAGER.get_browser_version_from_os(browser_type)
    )
    return not is_browser(ChromeType.GOOGLE) and is_browser(ChromeType.CHROMIUM)


def _to_manager(browser: str, *, root: Path) -> DriverManager:
    browser = browser.strip()
    manager_factory = AVAILABLE_DRIVERS.get(browser.lower())
    if not manager_factory:
        raise ValueError(
            f"Unsupported browser {browser!r} for webdriver download!"
            f" (choose from: {', '.join(SUPPORTED_BROWSERS.values())})"
        )

    if manager_factory == ChromeDriverManager and _is_chromium():
        manager_factory = functools.partial(
            manager_factory, chrome_type=ChromeType.CHROMIUM
        )
    cache_manager = DriverCacheManager(root_dir=str(root))
    manager = manager_factory(cache_manager=cache_manager)
    return manager


def _set_executable(path: str) -> None:
    st = os.stat(path)
    os.chmod(
        path,
        st.st_mode | stat.S_IXOTH | stat.S_IXGRP | stat.S_IEXEC,
    )


def download(browser: str, root: Path = DRIVER_ROOT) -> str:
    """Download a webdriver binary for the given browser and return the path to it."""
    manager = _to_manager(browser, root=root)
    with suppress_logging():
        path: str = manager.install()
    if platform.system() != "Windows":
        _set_executable(path)
    LOGGER.info("Downloaded webdriver to: %s", path)
    return path



from pathlib import Path
from typing import Dict, List, Optional

from RPA.core.vendor.deco import keyword as method
from RPA.core.windows.context import WindowsContext
from RPA.core.windows.helpers import IS_WINDOWS
from RPA.core.windows.locators import Locator, MatchObject, WindowsElement

if IS_WINDOWS:
    import uiautomation as auto
    from uiautomation.uiautomation import Control


StructureType = Dict[int, List[WindowsElement]]


class ElementMethods(WindowsContext):
    """Keywords for listing Windows GUI elements."""

    @staticmethod
    def _add_child_to_tree(
        control: "Control",
        structure: StructureType,
        *,
        locator: Optional[str],
        depth: int,
        path: str,
    ):
        # Adds current control child as element in the flattened tree structure.
        if locator and path:
            control_locator = f"{locator} > path:{path}"
        elif path:
            control_locator = f"path:{path}"
        else:
            control_locator = locator
        control.robocorp_click_offset = None
        element = WindowsElement(control, control_locator)
        structure.setdefault(depth, []).append(element)

    @method
    def print_tree(
        self,
        locator: Optional[Locator] = None,
        max_depth: int = 8,
        capture_image_folder: Optional[str] = None,
        log_as_warnings: Optional[bool] = False,
        return_structure: bool = False,
    ) -> Optional[StructureType]:
        # Cache how many brothers are in total given a child. (to know child position)
        brothers_count: Dict[int, int] = {}
        # Current path in the tree as children positions. (to compute the path locator)
        children_stack: List[int] = [-1] * (max_depth + 1)
        # Flattened tree of elements by depth level (to return the object if wanted).
        structure: StructureType = {}

        def get_children(ctrl: Control) -> List[Control]:
            children = ctrl.GetChildren()
            children_count = len(children)
            for child in children:
                brothers_count[hash(child)] = children_count
            return children

        target_elem = self.ctx.get_element(locator)
        locator: Optional[str] = WindowsElement.norm_locator(target_elem)
        root_ctrl = target_elem.item
        brothers_count[hash(root_ctrl)] = 1  # the root is always singular here

        image_idx = 1
        image_folder: Optional[Path] = None
        if capture_image_folder:
            image_folder = Path(capture_image_folder).expanduser().resolve()
            image_folder.mkdir(parents=True, exist_ok=True)

        control_log = {
            None: self.logger.debug,
            False: self.logger.info,
            True: self.logger.warning,
        }[log_as_warnings]

        for control, depth, children_remaining in auto.WalkTree(
            root_ctrl,
            getChildren=get_children,
            includeTop=True,
            maxDepth=max_depth,
        ):
            control_str = str(control)

            if image_folder:
                element = WindowsElement(control, locator)
                capture_filename = f"{control.ControlType}_{image_idx}.png"
                image_path = image_folder / capture_filename
                try:
                    self.ctx.screenshot(element, image_path)
                except Exception as exc:  # pylint: disable=broad-except
                    self.logger.warning(
                        "Couldn't capture into %r due to: %s", image_path, exc
                    )
                else:
                    control_str += f"    Image: {capture_filename}"
                image_idx += 1

            space = " " * depth * 4
            child_pos = brothers_count[hash(control)] - children_remaining
            children_stack[depth] = child_pos
            path = MatchObject.PATH_SEP.join(
                str(pos) for pos in children_stack[1 : depth + 1]
            )
            control_log(f"{space}{depth}-{child_pos}. {control_str}    Path: {path}")

            if return_structure:
                self._add_child_to_tree(
                    control,
                    structure,
                    locator=locator,
                    depth=depth,
                    path=path,
                )

        return structure if return_structure else None



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import inspect


def not_keyword(func):
    """Decorator to disable exposing functions or methods as keywords.

    Examples::

        @not_keyword
        def not_exposed_as_keyword():
            # ...

        def exposed_as_keyword():
            # ...

    Alternatively the automatic keyword discovery can be disabled with
    the :func:`library` decorator or by setting the ``ROBOT_AUTO_KEYWORDS``
    attribute to a false value.

    New in Robot Framework 3.2.
    """
    func.robot_not_keyword = True
    return func


not_keyword.robot_not_keyword = True


@not_keyword
def keyword(name=None, tags=(), types=()):
    """Decorator to set custom name, tags and argument types to keywords.

    This decorator creates ``robot_name``, ``robot_tags`` and ``robot_types``
    attributes on the decorated keyword function or method based on the
    provided arguments. Robot Framework checks them to determine the keyword's
    name, tags, and argument types, respectively.

    Name must be given as a string, tags as a list of strings, and types
    either as a dictionary mapping argument names to types or as a list
    of types mapped to arguments based on position. It is OK to specify types
    only to some arguments, and setting ``types`` to ``None`` disables type
    conversion altogether.

    If the automatic keyword discovery has been disabled with the
    :func:`library` decorator or by setting the ``ROBOT_AUTO_KEYWORDS``
    attribute to a false value, this decorator is needed to mark functions
    or methods keywords.

    Examples::

        @keyword
        def example():
            # ...

        @keyword('Login as user "${user}" with password "${password}"',
                 tags=['custom name', 'embedded arguments', 'tags'])
        def login(user, password):
            # ...

        @keyword(types={'length': int, 'case_insensitive': bool})
        def types_as_dict(length, case_insensitive):
            # ...

        @keyword(types=[int, bool])
        def types_as_list(length, case_insensitive):
            # ...

        @keyword(types=None])
        def no_conversion(length, case_insensitive=False):
            # ...
    """
    if inspect.isroutine(name):
        return keyword()(name)

    def decorator(func):
        func.robot_name = name
        func.robot_tags = tags
        func.robot_types = types
        return func

    return decorator


@not_keyword
def library(
    scope=None, version=None, doc_format=None, listener=None, auto_keywords=False
):
    """Class decorator to control keyword discovery and other library settings.

    By default disables automatic keyword detection by setting class attribute
    ``ROBOT_AUTO_KEYWORDS = False`` to the decorated library. In that mode
    only methods decorated explicitly with the :func:`keyword` decorator become
    keywords. If that is not desired, automatic keyword discovery can be
    enabled by using ``auto_keywords=True``.

    Arguments ``scope``, ``version``, ``doc_format`` and ``listener`` set the
    library scope, version, documentation format and listener by using class
    attributes ``ROBOT_LIBRARY_SCOPE``, ``ROBOT_LIBRARY_VERSION``,
    ``ROBOT_LIBRARY_DOC_FORMAT`` and ``ROBOT_LIBRARY_LISTENER``, respectively.
    These attributes are only set if the related arguments are given and they
    override possible existing attributes in the decorated class.

    Examples::

        @library
        class KeywordDiscovery:

            @keyword
            def do_something(self):
                # ...

            def not_keyword(self):
                # ...


        @library(scope='GLOBAL', version='3.2')
        class LibraryConfiguration:
            # ...

    The ``@library`` decorator is new in Robot Framework 3.2.
    """
    if inspect.isclass(scope):
        return library()(scope)

    def decorator(cls):
        if scope is not None:
            cls.ROBOT_LIBRARY_SCOPE = scope
        if version is not None:
            cls.ROBOT_LIBRARY_VERSION = version
        if doc_format is not None:
            cls.ROBOT_LIBRARY_DOC_FORMAT = doc_format
        if listener is not None:
            cls.ROBOT_LIBRARY_LISTENER = listener
        cls.ROBOT_AUTO_KEYWORDS = auto_keywords
        return cls

    return decorator



import logging
import warnings
from typing import Union, List

try:
    from robot.libraries.BuiltIn import BuiltIn
    from robot.running.context import EXECUTION_CONTEXTS
except ModuleNotFoundError:
    BuiltIn = None
    EXECUTION_CONTEXTS = None

from RPA.core.helpers import required_param


def deprecation(message):
    """Emits once a deprecation warning log given the provided `message`."""
    logging.captureWarnings(True)
    try:
        warnings.warn(message, FutureWarning, stacklevel=2)
    finally:
        logging.captureWarnings(False)


class RobotLogListener:
    """`RobotLogListener` is a library for controlling logging during
    a Robot Framework execution using the listener API.

    **About keyword parameters**

    Parameters `names` and `keywords` for keywords `Mute Run On Failure` and `Register Protected Keywords`
    do not need to be full names of keywords, ie. all keywords matching even partially will be affected.
    `Run Keyword` would match all `BuiltIn` library keywords (17 keywords in RF 3.2.1) and of course all
    `Run Keyword` named keywords in any resource and/or library file which are imported would be matched also.

    **Mute Run On Failure**

    This keyword is to be used specifically with `RPA.Browser.Selenium` library, which extends
    `SeleniumLibrary`.  Normally most of the `SeleniumLibrary` keywords execute `run_on_failure`
    behaviour, which can be set at library initialization. By default this behaviour is running
    `Capture Page Screenshot` keyword on failure.

    In the example task `Check the official website` below the keyword `Run Keyword` is muted and when
    it runs the keyword `Element Should Be Visible` then those failures do not create page screenshots
    into log file.

    It is also possible to change default failure behaviour by giving parameter
    `optional_keyword_to_run` for `Mute Run On Failure`, see task `Check the official website with error log`.
    This optional keyword would be then executed on failure. Due to the underlying `SeleniumLibrary`
    implementation this keyword can't have arguments.

    Example of using `Mute Run On Failure` without and with optional keyword to run.

    .. code-block:: robotframework

       *** Settings ***
       Library         RPA.Browser.Selenium
       Library         RPA.RobotLogListener
       Task Setup      Set Task Variable   ${TRIES}   1
       Task Teardown   Close All Browsers

       *** Tasks ***
       Check the official website
          Mute Run On Failure   Run Keyword
          Open Available Browser   https://www.robocorp.com
          Check for visible element
          Capture Page Screenshot

       Check the official website with error log
          Mute Run On Failure   Run Keyword  optional_keyword_to_run=Log tries
          Open Available Browser   https://www.robocorp.com
          Check for visible element
          Capture Page Screenshot

       *** Keywords ***
       Check for visible element
          FOR  ${idx}  IN RANGE  1   20
             Set Task Variable   ${TRIES}   ${idx}
             ${status}   Run Keyword And Return Status   Element Should Be Visible  id:xyz
             Exit For Loop If   '${status}' == 'PASS'
             Sleep  2s
          END

       Log tries
          Log  Checked element visibility ${TRIES} times

    **Register Protected Keywords**

    This keyword is used to totally disable logging for named keywords. In the example below
    the keyword `This will not output` is protected and it will not be logging into Robot Framework
    log files.

    **Robot Framework**

    .. code-block:: robotframework

        *** Settings ***
        Library         RPA.RobotLogListener

        *** Tasks ***
        Protecting keywords
           This will not output        # will output because called before register
           Register Protected Keywords    This will not output
           This will not output        # is now registered
           This will output

        *** Keywords ***
        This will not output
           Log   1

        This will output
           Log   2

    **Python**

    .. code-block:: python

        from robot.libraries.BuiltIn import BuiltIn, RobotNotRunningError
        from RPA.RobotLogListener import RobotLogListener

        try:
           BuiltIn().import_library("RPA.RobotLogListener")
        except RobotNotRunningError:
           pass

        class CustomLibrary:

           def __init__(self):
              listener = RobotLogListener()
              listener.register_protected_keywords(
                    ["CustomLibrary.special_keyword"]
              )

           def special_keyword(self):
              print('will not be written to log')
              return 'not shown in the log'
    """  # noqa: E501

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"
    ROBOT_LISTENER_API_VERSION = 2

    KEYWORDS_TO_PROTECT = ["rpa.robocloud.secrets.", "rpa.robocorp.vault."]
    KEYWORDS_TO_MUTE = []
    INFO_LEVEL_KEYWORDS = []

    def __init__(self) -> None:
        self.ROBOT_LIBRARY_LISTENER = self
        self.logger = logging.getLogger(__name__)
        self.stack = []

        # Run-on-failure
        self.muted_keyword = None
        self.muted_optionals = []
        self.muted_previous = {}

    def only_info_level(self, names: Union[str, List] = None):
        """Register keywords that are allowed only INFO level logging

        :param names: list of keywords to protect
        """
        required_param(names, "only_info_level")
        if not isinstance(names, list):
            names = [names]

        for name in names:
            normalized = self._normalize(name)
            if normalized not in self.INFO_LEVEL_KEYWORDS:
                self.INFO_LEVEL_KEYWORDS.append(normalized)

    def register_protected_keywords(self, names: Union[str, List] = None) -> None:
        """Register keywords that are not going to be logged into Robot Framework logs.

        :param names: list of keywords to protect
        """
        required_param(names, "register_protected_keywords")
        if not isinstance(names, list):
            names = [names]

        for name in names:
            normalized = self._normalize(name)
            if normalized not in self.KEYWORDS_TO_PROTECT:
                self.KEYWORDS_TO_PROTECT.append(normalized)

    def mute_run_on_failure(
        self, keywords: Union[str, List] = None, optional_keyword_to_run: str = None
    ) -> None:
        """Set keywords which should not execute `SeleniumLibrary`
        default behaviour of running keyword on failure.

        :param keywords: list of keywords to mute
        :param optional_keyword_to_run: name of the keyword to execute
            if keyword defined by `keywords` fail

        Keyword names do not need to be full names of keywords, ie. all keywords
        matching even partially will be affected. `Run Keyword` would match all
        `BuiltIn` library keywords (17 keywords in RF 3.2.1) and of course all
        `Run Keyword` named keywords in any resource and/or library file which
        are imported would be matched also.

        By default `SeleniumLibrary` executes `Capture Page Screenshot`
        on failure.

        If `optional_keyword_to_run` is not given then nothing is done
        on failure, but this can be set to override `SeleniumLibrary`
        default behaviour for a set of keywords.
        """
        if not self._is_robot_running():
            raise RuntimeError("Not supported outside Robot Framework")

        required_param(keywords, "mute_run_on_failure")
        if not isinstance(keywords, list):
            keywords = [keywords]

        for keyword in keywords:
            normalized = self._normalize(keyword)
            if normalized not in self.KEYWORDS_TO_MUTE:
                self.KEYWORDS_TO_MUTE.append(normalized)

        for library in ("RPA.Browser", "RPA.Browser.Selenium"):
            status, instance = BuiltIn().run_keyword_and_ignore_error(
                "get_library_instance", library
            )
            if status == "PASS":
                self.muted_optionals.append((instance, optional_keyword_to_run))

    def start_keyword(self, name, attributes):  # pylint: disable=W0613
        """Listener method for keyword start.

        :param name: keyword name
        :param attributes: keyword attributes

        If `name` exists in the protected keywords list then log level is
        temporarily set to NONE.
        """
        if not self._is_robot_running():
            return

        normalized = self._normalize(name)

        if any(name in normalized for name in self.KEYWORDS_TO_PROTECT):
            self.logger.info("Protecting keyword: %s", name)
            old = BuiltIn().set_log_level("NONE")
            self.stack.append((normalized, old))
        elif any(name in normalized for name in self.INFO_LEVEL_KEYWORDS):
            current = BuiltIn().get_variable_value("${LOG_LEVEL}")
            if current not in ("WARN", "ERROR", "NONE"):
                old = BuiltIn().set_log_level("INFO")
                self.stack.append((normalized, old))

        # Run-on-failure

        if self.muted_keyword:
            return

        previous = {}
        for library, optional in self.muted_optionals:
            if any(name in normalized for name in self.KEYWORDS_TO_MUTE):
                keyword = library.register_keyword_to_run_on_failure(optional)
                previous[library.__class__.__name__] = keyword

        if previous:
            self.logger.debug("Muting failures before keyword: %s", name)
            self.muted_keyword = normalized
            self.muted_previous = previous

    def end_keyword(self, name, attributes):  # pylint: disable=W0613
        """Listener method for keyword end.

        :param name: keyword name
        :param attributes: keyword attributes

        If `name` exists in the protected keywords list then log level is
        restored back to level it was before settings to NONE.
        """
        if not self._is_robot_running():
            return

        normalized = self._normalize(name)

        if self.stack and self.stack[-1][0] == normalized:
            _, old = self.stack.pop()
            BuiltIn().set_log_level(old)

        # Run-on-failure

        if normalized != self.muted_keyword:
            return

        self.logger.debug("Un-muting failures after keyword: %s", name)
        for library, _ in self.muted_optionals:
            previous = self.muted_previous.get(library.__class__.__name__)
            if previous:
                library.register_keyword_to_run_on_failure(previous)

        self.muted_keyword = None
        self.muted_previous = {}

    def _normalize(self, name: str) -> str:
        """Modifies keyword name for programmatic use.

        Keyword is lowercased and spaces are replaced by underscores.

        :param name: keyword name to normalize
        :return: normalized keyword
        """
        return name.lower().replace(" ", "_")

    def _is_robot_running(self):
        return BuiltIn is not None and EXECUTION_CONTEXTS.current is not None
